;;;; AJL - geneticky algoritmus II (ohranicenia), v5
;;; (C) MM, 2009

;;;; Zadanie: Zbieranie pokladu  
;;;
;;;  K dispozicii je obdlznikova plocha na ktorej kazde policko
;;;  obsahuje urcite mnozstvo pokladu. Cielom je najst taku
;;;  cestu z jedneho rohu do ineho, aby mnozstvo pokladu pozdlz
;;;  tejto cesty bolo co najvacsie.
;;;  Vychodzi bod je v lavom hornom rohu a koncovy v pravom
;;;  dolnom rohu. Povolene kroky su iba posun doprava alebo dole
;;;  o jednu poziciu.

;;;; Verzia 1 ******************************************************

(defvar *field-x*)
(defvar *field-y*)

(defvar *path-length*)

(defun new-problem (&optional (x 5) (y 5))
  (setf *field-x* x)
  (setf *field-y* y)
  (setf *path-length* (+ x y -2)))

(defun generate-path ()
  (let ((body (make-array *path-length*)))
    (dotimes (i *path-length* body)
      (setf (aref body i) (random 2)))))

(defun valid-path-p (path)
  (let ((sum 0))
    (dotimes (i *path-length*)
      (incf sum (aref path i)))
    (if (= sum (1- *field-x*)) t nil)))

(defun repair-path (path)
  (let ((sum0 0) (sum1 0))
    (dotimes (i *path-length*)
      (if (= (aref path i) 0)
	  (if (< sum0 (1- *field-y*))
	      (incf sum0)
	      (setf (aref path i) 1))
	  (if (< sum1 (1- *field-x*))
	      (incf sum1)
	      (setf (aref path i) 0))))
    path))

(defun fitness (path)
  (let ((sum 0)
	(x 0) (y 0))
    (dotimes (i *path-length*)
      (if (= (aref path i) 0) (incf y) (incf x))
      (incf sum (max x y)))
    sum))

;;;; Verzia 2 ******************************************************

(defun repair-path (path-orig)
  (let ((sum0 0) (sum1 0)
	(path (copy-seq path-orig)))
    (dotimes (i *path-length*)
      (if (= (aref path i) 0)
	  (if (< sum0 (1- *field-y*))
	      (incf sum0)
	      (setf (aref path i) 1))
	  (if (< sum1 (1- *field-x*))
	      (incf sum1)
	      (setf (aref path i) 0))))
    path))

(defun print-path (path)
  (let ((field (make-array (list *field-x* *field-y*) :initial-element 0)))
    (setf (aref field 0 0) 1)
    (let ((x 0) (y 0))
      (dotimes (i *path-length*)
	(if (= (aref path i) 0) (incf y) (incf x))
	(and (< x *field-x*) (< y *field-y*)
	     (setf (aref field x y) 1))))
    (dotimes (y *field-y*)
      (dotimes (x *field-x*)
	(format t "~[ .~; *~]" (aref field x y)))
      (format t "~%"))))

;;;; Verzia 3 ******************************************************

(defclass fitness ()
  ((fitness-value
    :initform nil
    :accessor fitness-value)
   (max-fitness
    :initform -1
    :accessor max-fitness
    :allocation :class)
   (best-so-far
    :accessor fitness-best-so-far
    :allocation :class)))

(defmethod initialize-instance :after ((repre fitness) &key)
  (setf (slot-value repre 'max-fitness) 0))

(defclass repaired-fitness (fitness)
  ((repaired-fitness
    :initform nil
    :accessor repaired-fitness)))

(defclass representation ()
  ((body-structure
    :initarg :body-structure
    :initform (error "Must supply a body structure.")
    :reader body-structure)))

(defclass repaired-representation (representation)
  ((repaired-structure
    :reader repaired-structure)))

(defmethod initialize-instance :after ((repre repaired-representation) &key)
  (setf (slot-value repre 'repaired-structure)
	(repair-path (slot-value repre 'body-structure))))


(defclass individual (representation fitness) ())

(defclass repaired-individual (repaired-representation repaired-fitness) ())

(defclass invalid-individual (representation repaired-fitness) ())

;;;; Verzia 4 ******************************************************

(defparameter *prob-repair* 0.3)

(defun make-individual (body)
  (if (valid-path-p body)
      (make-instance 'individual :body-structure body)
      (if (< (random 1.0) *prob-repair*)
	  (make-instance 'repaired-individual :body-structure body)
	  (make-instance 'invalid-individual :body-structure body))))

(defmethod print-object ((object repaired-individual) stream)
  (print-unreadable-object (object stream :type t)
    (with-slots (body-structure repaired-structure repaired-fitness
				max-fitness)
	object
      (format stream "~%path: ~a ~%rep-path: ~a ~%rep-fitness: ~a ~%max: ~a"
	      body-structure repaired-structure repaired-fitness
	      max-fitness))))

(defmethod print-object ((object invalid-individual) stream)
  (print-unreadable-object (object stream :type t)
    (with-slots (body-structure repaired-fitness max-fitness)
	object
      (format stream "~%path: ~a ~%rep-fitness: ~a ~%max: ~a"
	      body-structure repaired-fitness max-fitness))))

(defmethod print-object ((object individual) stream)
  (print-unreadable-object (object stream :type t)
    (with-slots (body-structure fitness-value max-fitness)
	object
      (format stream "~%path: ~a ~%fitness: ~a ~%max: ~a"
	      body-structure fitness-value max-fitness))))

;;;; Verzia 5 ******************************************************

(defgeneric evaluate (individual))

(defmethod evaluate :before ((individual repaired-individual))
  (setf (repaired-fitness individual)
	(fitness (repaired-structure individual))))

(defmethod evaluate :before ((individual invalid-individual))
  (setf (repaired-fitness individual)
	(fitness (repair-path (body-structure individual)))))

(defmethod evaluate ((individual repaired-fitness))
  (setf (fitness-value individual)
	(repaired-fitness individual)))

(defmethod evaluate ((individual individual))
  (setf (fitness-value individual)
	(fitness (body-structure individual))))

(defmethod evaluate :after ((individual fitness))
  (when (> (fitness-value individual) (max-fitness individual))
    (setf (max-fitness individual) (fitness-value individual))))

;;;; Verzia 6 ******************************************************

(defparameter *prob-crossover* 0.6)

(defun cross (xorig yorig)
  (let ((swap)
	(x (copy-seq xorig)) (y (copy-seq yorig)))
    (when (< (random 1.0) *prob-crossover*)
      (dotimes (i *path-length*)
	(when (< (random 1.0) 0.2)
	  (setf swap (aref x i))
	  (setf (aref x i) (aref y i))
	  (setf (aref y i) swap))))
    (values x y)))


(defparameter *prob-mutation* 0.05)

(defun mutate (body-orig)
  (let ((body (copy-seq body-orig)))
    (dotimes (i *path-length* body)
      (when (< (random 1.0) *prob-mutation*)
	(if (= (aref body i) 0)
	    (setf (aref body i) 1)
	    (setf (aref body i) 0))))))

;;;; Verzia 7 ******************************************************

(defparameter *population-size* 4)

(defun generate ()
  (let ((population (make-array *population-size*)))
    (dotimes (i *population-size*)
      (setf (aref population i) (make-individual (generate-path))))
    population))

(defun distribution (population)
  (dotimes (i *population-size* population)
    (evaluate (aref population i))))

(defparameter *t* 0.5)

(defun selectone (sorted-population)
  (let ((considered (floor (* *population-size* *t*))))
    (aref sorted-population (random considered))))

(defgeneric crossover (ind1 ind2))

(defmethod crossover ((ind1 repaired-individual) (ind2 repaired-individual))
  (cross (repaired-structure ind1)
	 (repaired-structure ind2)))

(defmethod crossover ((ind1 repaired-individual) ind2)
  (cross (repaired-structure ind1)
	 (body-structure ind2)))

(defmethod crossover (ind1 (ind2 repaired-individual))
  (crossover ind2 ind1))

(defmethod crossover (ind1 ind2)
  (cross (body-structure ind1)
	 (body-structure ind2)))

(defun reproduce (population)
  (let* ((offspring (make-array (length population) :fill-pointer 0)))
    (setf population
	  (sort population #'> :key #'(lambda (x) (fitness-value x))))
    (dotimes (i (floor (/ (length population) 2)))
      (multiple-value-bind (x y)
	  (crossover (selectone population) (selectone population))
	(vector-push (make-individual (mutate x)) offspring)
	(vector-push (make-individual (mutate y)) offspring)))
    offspring))

(defparameter *max-gen* 30)

(defun ga-run ()
  (let ((population (distribution (generate))))
    (dotimes (i *max-gen*)                     
      (setf population (distribution (reproduce population))))
    population))

;;;; Verzia 8 ******************************************************

(defun reproduce (population)
  (let* ((offspring (make-array (length population) :fill-pointer 0
				:adjustable t)))
    (setf population
	  (sort population #'> :key #'(lambda (x) (fitness-value x))))
    (dotimes (i (floor (/ (length population) 2)))
      (multiple-value-bind (x y)
	  (crossover (selectone population) (selectone population))
	(vector-push-extend (make-individual (mutate x)) offspring)
	(vector-push-extend (make-individual (mutate y)) offspring)))
    (vector-push-extend (aref population 0) offspring)
    (when (< *population-size* (length offspring)) (incf *population-size*))
    offspring))

;;;; Verzia 9 ******************************************************

(defun generate ()
  (let ((population (make-array *population-size*)))
    (dotimes (i *population-size*)
      (setf (aref population i) (make-individual (generate-path))))
    (setf (fitness-best-so-far (aref population 0)) 0)
    population))

(defun elite (population)
  (aref population
	(position-if #'(lambda (x) (= (fitness-value x) (max-fitness x)))
		     population)))

(defun ga-run ()
  (let ((population (distribution (generate))))
    (dotimes (i *max-gen*)                     
      (setf population (distribution (reproduce population)))
      (print-best i (elite population)))))

(defun print-best (generation best)
  (with-accessors ((so-far fitness-best-so-far)
		   (fitness fitness-value))
    best
    (when (< so-far fitness)
      (setf so-far fitness)
      (let ((path (cond ((typep best 'repaired-representation)
			 (repaired-structure best))
			((typep best 'invalid-individual)
			 (repair-path (body-structure best)))
			(t (body-structure best)))))
	(format t "generation: ~a fitness: ~a~%" generation fitness)
	(print-path path)))))


(defun new-alg (&key (pr 0.3) (pc 0.6) (pm 0.05) (ps 4) (tr 0.5) (mg 30))
  (setf *prob-repair* pr)
  (setf *prob-crossover* pc)
  (setf *prob-mutation* pm)
  (setf *population-size* ps)
  (setf *t* tr)
  (setf *max-gen* mg))









